# std
from typing import Dict, List, Union

# external
from prodict import Prodict

# type definitions
AtomicData = Union[type(None), bool, int, str]
ListData = List[Union[AtomicData, 'ListData', 'DictData']]
_DictData = Dict[str, Union[AtomicData, ListData, 'DictData']]
_ProdictData = Prodict[str, Union[AtomicData, ListData, 'DictData']]
DictData = Union[_DictData, _ProdictData]
Data = Union[AtomicData, ListData, DictData]



# internal
from laz.plugins.plugin import Plugin


class TerraformPlugin(Plugin):

    @property
    def config(self) -> dict:
        return self.context.data.get('terraform', {})

    @property
    def hooks(self) -> dict:
        return self.context.data.get('hooks', {})

    def before_target(self):
        self._handle_log()
        self._handle_log_path()
        self._handle_workspace()
        self._handle_tf_vars()
        self._handle_hooks()

    def _handle_log(self):
        tf_log = self.config.get('log')
        if tf_log:
            self.context.push({'env': {'TF_LOG': tf_log}})

    def _handle_log_path(self):
        tf_log_path = self.config.get('log_path')
        if tf_log_path:
            self.context.push({'env': {'TF_LOG_PATH': tf_log_path}})

    def _handle_workspace(self):
        tf_workspace = self.config.get('workspace')
        if tf_workspace:
            self.context.push({'env': {'TF_WORKSPACE': tf_workspace}})

    def _handle_tf_vars(self):
        tf_vars = self.config.get('tf_vars', {})
        env = {f'TF_VAR_{key}': val for key, val in tf_vars.items()}
        if env:
            self.context.push({'env': env})

    def _handle_hooks(self):
        if self.context.data['args'][0] == 'terraform':
            self._handle_terraform_hook('before')
            self._handle_terraform_hook('after')
            self._handle_terraform_subcommand_hook('before')
            self._handle_terraform_subcommand_hook('after')

    def _handle_terraform_hook(self, before_or_after: str = 'before'):
        hook_name = f'{before_or_after}_terraform'
        if hook_name in self.hooks:
            hook = self.hooks[hook_name]
            if not isinstance(hook, list):
                hook = [hook]
            data = {'hooks': {f'{before_or_after}_target': hook}}
            self.context.push(data)

    def _handle_terraform_subcommand_hook(self, before_or_after: str = 'before'):
        subcommand = self.context.data['args'][1]
        hook_name = f'{before_or_after}_terraform_{subcommand}'
        if hook_name in self.hooks:
            hook = self.hooks[hook_name]
            if not isinstance(hook, list):
                hook = [hook]
            data = {'hooks': {f'{before_or_after}_target': hook}}
            self.context.push(data)



# std
from typing import Callable, Dict, List, Optional as Opt

# internal
from laz.utils.errors import LazTypeError
from laz.utils.types import Data, AtomicData, DictData, ListData


class Walk:

    def __init__(
            self,
            data: Data,
            context: Opt[DictData] = None,
            atomic_handler: Opt[Callable[[AtomicData, Opt[DictData]], Data]] = None,
            dict_handler: Opt[Callable[[Dict, Opt[DictData]], Data]] = None,
            list_handler: Opt[Callable[[List, Opt[DictData]], Data]] = None,
    ):
        self.data = data
        self.context = context
        self.atomic_handler = atomic_handler
        self.dict_handler = dict_handler
        self.list_handler = list_handler

    def walk(self):
        return self._walk(self.data)

    def _walk(self, data: Data):
        if isinstance(data, dict):
            return self._walk_dict(data)
        elif isinstance(data, list):
            return self._walk_list(data)
        elif isinstance(data, (type(None), bool, int, str)):
            return self._walk_atomic(data)
        else:
            raise LazTypeError(f'Cannot walk input of type {type(data)}')

    def _walk_dict(self, data: DictData):
        for key in data.keys():
            data[key] = self._walk(data[key])
        if self.dict_handler is not None:
            return self.dict_handler(data, self.context)
        return data

    def _walk_list(self, data: ListData):
        for i in range(len(data)):
            data[i] = self._walk(data[i])
        if self.list_handler is not None:
            return self.list_handler(data, self.context)
        return data

    def _walk_atomic(self, data: AtomicData):
        if self.atomic_handler is not None:
            return self.atomic_handler(data, self.context)
        return data



# std
import os
from typing import Dict, Union

# internal
from laz.utils.errors import LazValueError, LazError
from laz.utils import log
from laz.utils.types import Data
from laz.utils.contexts import with_environ
from laz.model.action import Action
from laz.model.configuration import Configuration
from laz.model.target import Target
from laz.plugins.plugin import Plugin


class AwsPlugin(Plugin):

    def before_all(self):
        env = {}
        aws_profile = self.context.data.get('aws', {}).get('profile')
        if aws_profile is not None:
            env['AWS_PROFILE'] = aws_profile
        aws_region = self.context.data.get('aws', {}).get('region')
        if aws_region is not None:
            env['AWS_DEFAULT_REGION'] = aws_region
        if env:
            self.context.push({'env': env})


class AwsAction(Action):

    def run(self):
        if isinstance(self.run_data['aws'], dict) and 'assume_role' in self.run_data['aws']:
            return self._assume_role()
        elif isinstance(self.run_data['aws'], dict):
            return self._arbitrary_aws_action()
        else:
            raise LazValueError(f'Invalid aws plugin action')

    def _assume_role(self):
        import boto3
        aws: dict = self.run_data['aws']
        kwargs: Dict[str, str] = aws['assume_role']
        if 'RoleSessionName' not in kwargs:
            kwargs['RoleSessionName'] = os.environ['USER']
        with with_environ(self.context.data.get('env', {})):
            sts = boto3.client('sts')
            response = sts.assume_role(**kwargs)
        status_code = response['ResponseMetadata']['HTTPStatusCode']
        if status_code >= 300:
            raise LazError(f'Assume Role Error: HTTPStatusCode {status_code}')
        credentials = response['Credentials']
        env = {
            'AWS_ACCESS_KEY_ID': credentials['AccessKeyId'],
            'AWS_SECRET_ACCESS_KEY': credentials['SecretAccessKey'],
            'AWS_SESSION_TOKEN': credentials['SessionToken'],
        }
        self.context.push({'env': env})

    def _arbitrary_aws_action(self):
        import boto3
        with with_environ(self.context.data.get('env') or {}):
            aws: Dict[str, Dict[str, Dict[str, str]]] = self.run_data['aws']
            service: str = list(aws.keys())[0]
            subcommand: str = list(aws[service].keys())[0]
            kwargs: Dict[str, str] = aws[service][subcommand] or {}
            client = boto3.client(service)
            method = getattr(client, subcommand)
            response = method(**kwargs)
        status_code = response['ResponseMetadata']['HTTPStatusCode']
        if status_code >= 300:
            log.error(str(aws))
            log.error(str(response))
            raise LazError(f'AWS action failed')
        return response

    @classmethod
    def is_handler(cls, context: Union[Configuration, Target], run_data: Data) -> bool:
        return isinstance(run_data, dict) and 'aws' in run_data



# std
from contextlib import contextmanager
from copy import deepcopy
import os
from typing import Dict

# internal
from laz.utils.merge import merge


@contextmanager
def in_dir(dirpath: str):
    previous_dirpath = os.getcwd()
    try:
        os.chdir(dirpath)
        yield
    finally:
        os.chdir(previous_dirpath)


@contextmanager
def with_environ(env: Dict[str, str]):
    original = deepcopy(os.environ)
    try:
        os.environ = merge(dict(os.environ), env)
        yield
    finally:
        os.environ = original



# std
from __future__ import annotations
from copy import deepcopy
import os
from typing import List

# external
import yaml
from yaml.loader import SafeLoader

# internal
from laz.model.base import BaseObject
from laz.model.target import Target
from laz.utils.prodict import prodictify
from laz.utils.types import Data


class Configuration(BaseObject):

    def __init__(self, id: str, **data: Data):
        super().__init__(id, **deepcopy(DEFAULT_CONFIGURATION))
        self.push(data)

    @property
    def filepath(self):
        return self.id

    @property
    def name(self):
        return os.path.basename(os.path.dirname(self.filepath))

    @property
    def target_names(self) -> List[str]:
        return list(self.data.get('targets', {}).keys())

    def get_target(self, name: str) -> Target:
        return Target(name, **self.data.get('targets', {}).get(name, {}) or {})

    @classmethod
    def deserialize(cls, id: str, serialized: str) -> Configuration:
        data = prodictify(yaml.load(serialized, Loader=SafeLoader) or {})
        return cls(id, **data)

    @classmethod
    def load(cls, filepath: str) -> Configuration:
        with open(filepath, 'r') as fh:
            return Configuration.deserialize(filepath, fh.read())


DEFAULT_CONFIGURATION = {
    'laz': {
        'default_action': 'default',
        'default_target': 'default',
        'error_on_no_targets': False,
    },
    'env': {},
    'targets': {},
    'actions': {},
}



# std
from copy import deepcopy
from typing import Optional as Opt

# external
from prodict import Prodict

# internal
from laz.utils.errors import LazTypeError
from laz.utils.types import AtomicData, Data, DictData, ListData
from laz.plugins.plugin import Plugin
from laz.utils.walk import Walk
from laz.utils.prodict import prodictify


class AliasPlugin(Plugin):

    def before_target(self):
        aliased = alias(self.context.data)
        self.context.replace(aliased)


def alias(data: DictData) -> Data:
    data = deepcopy(data)
    return _alias(data, data)


def _alias(data: DictData, context: DictData) -> Data:
    walk = Walk(data, context, atomic_handler=_handle_atomic)
    return walk.walk()


def _handle_atomic(data: AtomicData, context: DictData) -> Data:
    if isinstance(data, str):
        if data.startswith('='):
            for key, val in context.items():
                locals()[key] = val
            return _alias(eval(data[1:]), context)
    return data
